@@ -1,5 +1,5 @@
 /*
- * @(#) $Id: commands.c,v 1.13 2002/06/27 16:22:50 _cvs_pont Exp $
+ * @(#) $Id: commands.c,v 1.17 2004/11/17 10:59:24 nisse Exp $
  *
  * commands.c
  */
@@ -24,6 +24,10 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 #include "commands.h"
 
 int ls_longlist = 1;
@@ -893,7 +897,7 @@ get_file_or_folder(const char *arg, const char *command,
   const char* realdest;
   char *tmp;
 
-  struct stat st;
+  struct sftp_attrib attrib;
   int id;
   int mode = 0700;
   int mask;
@@ -931,15 +935,17 @@ get_file_or_folder(const char *arg, const char *command,
   
   /* FIXME: What do we expect filename_part( "foobar/" ) to return? */
   
-  id = lsftp_do_stat( arg, &st );
+  id = lsftp_do_stat( arg, &attrib );
   
   if( id > 0) /* Not a failure? */
     {
       lsftp_await_command( id );
-      mode = st.st_mode;
 
-      if( -1 == mode )
-	mode = 0700;       /* FIXME: Default mode */
+      if (attrib.flags & SSH_FILEXFER_ATTR_PERMISSIONS)
+	mode = attrib.permissions;
+      else
+	/* FIXME: Default mode */
+	mode = 0700;       
     }
 
   if( destname )              /* If given, use it */
@@ -1423,7 +1429,7 @@ int com_chown(const char *arg, const char *command)
   int gotuid = 0;
   int enough_parameters = 0;
   long newuid = 0;
-  struct stat st;
+  struct sftp_attrib attrib;
 
   if( !lsftp_connected() )   /* Bail out if not connected */
     return com_disconnected();
@@ -1435,7 +1441,7 @@ int com_chown(const char *arg, const char *command)
     {
       if( !gotuid ) 
 	{
-	  if( isdigit( tmp[0] ) ) /* Numeric GID? FIXME: Better check? */
+	  if( isdigit( (unsigned char) tmp[0] ) ) /* Numeric GID? FIXME: Better check? */
 	    newuid = atol( tmp );
 	  else
 	    {
@@ -1471,12 +1477,15 @@ int com_chown(const char *arg, const char *command)
 	
 	  while( (ptr = *glob++) )
 	    {
-	      id = lsftp_do_stat( ptr, &st );
-
+	      id = lsftp_do_stat( ptr, &attrib );
 	      if( id > 0) /* Not a failure? */
 		{
 		  lsftp_await_command( id );
-		  lsftp_do_chown( ptr, newuid, st.st_gid, command );
+
+		  if (! (attrib.flags & SSH_FILEXFER_ATTR_UIDGID))
+		    printf("Remote system doesn't support uids?\n");
+		  
+		  lsftp_do_chown(ptr, newuid, attrib.gid, command );
 		}
 	    }
 	
@@ -1502,7 +1511,7 @@ int com_chgrp(const char *arg, const char *command)
   int gotgid = 0;
   long newgid = 0;
   int enough_parameters = 0;
-  struct stat st;
+  struct sftp_attrib attrib;
 
   if( !lsftp_connected() )   /* Bail out if not connected */
     return com_disconnected();
@@ -1514,7 +1523,7 @@ int com_chgrp(const char *arg, const char *command)
     {
       if( !gotgid )  /* First time? */
 	{
-	  if( isdigit( tmp[0] ) ) /* Numeric GID? FIXME: Better check? */
+	  if( isdigit( (unsigned char) tmp[0] ) ) /* Numeric GID? FIXME: Better check? */
 	    newgid = atol( tmp );
 	  else
 	    {
@@ -1540,13 +1549,15 @@ int com_chgrp(const char *arg, const char *command)
 	  int id;
 
 	  enough_parameters++;
-	  id = lsftp_do_stat( tmp, &st );
+	  id = lsftp_do_stat( tmp, &attrib );
 
 	  if( id > 0) /* Not a failure? */
 	    {
 	      lsftp_await_command( id );
-	  
-	      lsftp_do_chown( tmp, st.st_uid, newgid, command );
+
+	      if (! (attrib.flags & SSH_FILEXFER_ATTR_UIDGID))
+		printf("Remote system doesn't support gids?\n");
+	      lsftp_do_chown( tmp, attrib.uid, newgid, command );
 	    }
 	}
 
@@ -1909,8 +1920,9 @@ uidstring( int uid )
   if( p ) /* Negative result */
     return p->pw_name;
 
-  snprintf( uids, 9, "%d", uid );
-
+  snprintf( uids, sizeof(uids), "%d", uid );
+  uids[sizeof(uids)-1] = 0;
+  
   return uids;
 }
 
@@ -1923,7 +1935,9 @@ gidstring( int gid )
   if( g )
     return g->gr_name;
 
-  snprintf( gids, 9, "%d", gid );
+  snprintf( gids, sizeof(gids), "%d", gid );
+  gids[sizeof(gids)-1] = 0;
+
   return gids;
 }
 
